# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nrpc/nrpc.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='nrpc/nrpc.proto',
  package='nrpc',
  syntax='proto3',
  serialized_pb=_b('\n\x0fnrpc/nrpc.proto\x12\x04nrpc\x1a google/protobuf/descriptor.proto\"\x86\x01\n\x05\x45rror\x12\x1e\n\x04type\x18\x01 \x01(\x0e\x32\x10.nrpc.Error.Type\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x10\n\x08msgCount\x18\x03 \x01(\r\":\n\x04Type\x12\n\n\x06\x43LIENT\x10\x00\x12\n\n\x06SERVER\x10\x01\x12\x07\n\x03\x45OS\x10\x03\x12\x11\n\rSERVERTOOBUSY\x10\x04\"\x06\n\x04Void\"\x0b\n\tNoRequest\"\t\n\x07NoReply\"\x1d\n\tHeartBeat\x12\x10\n\x08lastbeat\x18\x01 \x01(\x08*$\n\x0bSubjectRule\x12\x08\n\x04\x43OPY\x10\x00\x12\x0b\n\x07TOLOWER\x10\x01:6\n\x0epackageSubject\x12\x1c.google.protobuf.FileOptions\x18\xd0\x86\x03 \x01(\t:<\n\x14packageSubjectParams\x12\x1c.google.protobuf.FileOptions\x18\xd1\x86\x03 \x03(\t:M\n\x12serviceSubjectRule\x12\x1c.google.protobuf.FileOptions\x18\xd2\x86\x03 \x01(\x0e\x32\x11.nrpc.SubjectRule:L\n\x11methodSubjectRule\x12\x1c.google.protobuf.FileOptions\x18\xd3\x86\x03 \x01(\x0e\x32\x11.nrpc.SubjectRule:9\n\x0eserviceSubject\x12\x1f.google.protobuf.ServiceOptions\x18\xb8\x8e\x03 \x01(\t:?\n\x14serviceSubjectParams\x12\x1f.google.protobuf.ServiceOptions\x18\xb9\x8e\x03 \x03(\t:7\n\rmethodSubject\x12\x1e.google.protobuf.MethodOptions\x18\xa0\x96\x03 \x01(\t:=\n\x13methodSubjectParams\x12\x1e.google.protobuf.MethodOptions\x18\xa1\x96\x03 \x03(\t:7\n\rstreamedReply\x12\x1e.google.protobuf.MethodOptions\x18\xa2\x96\x03 \x01(\x08\x42\x1aZ\x18github.com/nats-rpc/nrpcb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_SUBJECTRULE = _descriptor.EnumDescriptor(
  name='SubjectRule',
  full_name='nrpc.SubjectRule',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='COPY', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TOLOWER', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=259,
  serialized_end=295,
)
_sym_db.RegisterEnumDescriptor(_SUBJECTRULE)

SubjectRule = enum_type_wrapper.EnumTypeWrapper(_SUBJECTRULE)
COPY = 0
TOLOWER = 1

PACKAGESUBJECT_FIELD_NUMBER = 50000
packageSubject = _descriptor.FieldDescriptor(
  name='packageSubject', full_name='nrpc.packageSubject', index=0,
  number=50000, type=9, cpp_type=9, label=1,
  has_default_value=False, default_value=_b("").decode('utf-8'),
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  options=None)
PACKAGESUBJECTPARAMS_FIELD_NUMBER = 50001
packageSubjectParams = _descriptor.FieldDescriptor(
  name='packageSubjectParams', full_name='nrpc.packageSubjectParams', index=1,
  number=50001, type=9, cpp_type=9, label=3,
  has_default_value=False, default_value=[],
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  options=None)
SERVICESUBJECTRULE_FIELD_NUMBER = 50002
serviceSubjectRule = _descriptor.FieldDescriptor(
  name='serviceSubjectRule', full_name='nrpc.serviceSubjectRule', index=2,
  number=50002, type=14, cpp_type=8, label=1,
  has_default_value=False, default_value=0,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  options=None)
METHODSUBJECTRULE_FIELD_NUMBER = 50003
methodSubjectRule = _descriptor.FieldDescriptor(
  name='methodSubjectRule', full_name='nrpc.methodSubjectRule', index=3,
  number=50003, type=14, cpp_type=8, label=1,
  has_default_value=False, default_value=0,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  options=None)
SERVICESUBJECT_FIELD_NUMBER = 51000
serviceSubject = _descriptor.FieldDescriptor(
  name='serviceSubject', full_name='nrpc.serviceSubject', index=4,
  number=51000, type=9, cpp_type=9, label=1,
  has_default_value=False, default_value=_b("").decode('utf-8'),
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  options=None)
SERVICESUBJECTPARAMS_FIELD_NUMBER = 51001
serviceSubjectParams = _descriptor.FieldDescriptor(
  name='serviceSubjectParams', full_name='nrpc.serviceSubjectParams', index=5,
  number=51001, type=9, cpp_type=9, label=3,
  has_default_value=False, default_value=[],
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  options=None)
METHODSUBJECT_FIELD_NUMBER = 52000
methodSubject = _descriptor.FieldDescriptor(
  name='methodSubject', full_name='nrpc.methodSubject', index=6,
  number=52000, type=9, cpp_type=9, label=1,
  has_default_value=False, default_value=_b("").decode('utf-8'),
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  options=None)
METHODSUBJECTPARAMS_FIELD_NUMBER = 52001
methodSubjectParams = _descriptor.FieldDescriptor(
  name='methodSubjectParams', full_name='nrpc.methodSubjectParams', index=7,
  number=52001, type=9, cpp_type=9, label=3,
  has_default_value=False, default_value=[],
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  options=None)
STREAMEDREPLY_FIELD_NUMBER = 52002
streamedReply = _descriptor.FieldDescriptor(
  name='streamedReply', full_name='nrpc.streamedReply', index=8,
  number=52002, type=8, cpp_type=7, label=1,
  has_default_value=False, default_value=False,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  options=None)

_ERROR_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='nrpc.Error.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CLIENT', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SERVER', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EOS', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SERVERTOOBUSY', index=3, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=136,
  serialized_end=194,
)
_sym_db.RegisterEnumDescriptor(_ERROR_TYPE)


_ERROR = _descriptor.Descriptor(
  name='Error',
  full_name='nrpc.Error',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='nrpc.Error.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='message', full_name='nrpc.Error.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='msgCount', full_name='nrpc.Error.msgCount', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ERROR_TYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=60,
  serialized_end=194,
)


_VOID = _descriptor.Descriptor(
  name='Void',
  full_name='nrpc.Void',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=196,
  serialized_end=202,
)


_NOREQUEST = _descriptor.Descriptor(
  name='NoRequest',
  full_name='nrpc.NoRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=204,
  serialized_end=215,
)


_NOREPLY = _descriptor.Descriptor(
  name='NoReply',
  full_name='nrpc.NoReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=217,
  serialized_end=226,
)


_HEARTBEAT = _descriptor.Descriptor(
  name='HeartBeat',
  full_name='nrpc.HeartBeat',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='lastbeat', full_name='nrpc.HeartBeat.lastbeat', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=228,
  serialized_end=257,
)

_ERROR.fields_by_name['type'].enum_type = _ERROR_TYPE
_ERROR_TYPE.containing_type = _ERROR
DESCRIPTOR.message_types_by_name['Error'] = _ERROR
DESCRIPTOR.message_types_by_name['Void'] = _VOID
DESCRIPTOR.message_types_by_name['NoRequest'] = _NOREQUEST
DESCRIPTOR.message_types_by_name['NoReply'] = _NOREPLY
DESCRIPTOR.message_types_by_name['HeartBeat'] = _HEARTBEAT
DESCRIPTOR.enum_types_by_name['SubjectRule'] = _SUBJECTRULE
DESCRIPTOR.extensions_by_name['packageSubject'] = packageSubject
DESCRIPTOR.extensions_by_name['packageSubjectParams'] = packageSubjectParams
DESCRIPTOR.extensions_by_name['serviceSubjectRule'] = serviceSubjectRule
DESCRIPTOR.extensions_by_name['methodSubjectRule'] = methodSubjectRule
DESCRIPTOR.extensions_by_name['serviceSubject'] = serviceSubject
DESCRIPTOR.extensions_by_name['serviceSubjectParams'] = serviceSubjectParams
DESCRIPTOR.extensions_by_name['methodSubject'] = methodSubject
DESCRIPTOR.extensions_by_name['methodSubjectParams'] = methodSubjectParams
DESCRIPTOR.extensions_by_name['streamedReply'] = streamedReply

Error = _reflection.GeneratedProtocolMessageType('Error', (_message.Message,), dict(
  DESCRIPTOR = _ERROR,
  __module__ = 'nrpc.nrpc_pb2'
  # @@protoc_insertion_point(class_scope:nrpc.Error)
  ))
_sym_db.RegisterMessage(Error)

Void = _reflection.GeneratedProtocolMessageType('Void', (_message.Message,), dict(
  DESCRIPTOR = _VOID,
  __module__ = 'nrpc.nrpc_pb2'
  # @@protoc_insertion_point(class_scope:nrpc.Void)
  ))
_sym_db.RegisterMessage(Void)

NoRequest = _reflection.GeneratedProtocolMessageType('NoRequest', (_message.Message,), dict(
  DESCRIPTOR = _NOREQUEST,
  __module__ = 'nrpc.nrpc_pb2'
  # @@protoc_insertion_point(class_scope:nrpc.NoRequest)
  ))
_sym_db.RegisterMessage(NoRequest)

NoReply = _reflection.GeneratedProtocolMessageType('NoReply', (_message.Message,), dict(
  DESCRIPTOR = _NOREPLY,
  __module__ = 'nrpc.nrpc_pb2'
  # @@protoc_insertion_point(class_scope:nrpc.NoReply)
  ))
_sym_db.RegisterMessage(NoReply)

HeartBeat = _reflection.GeneratedProtocolMessageType('HeartBeat', (_message.Message,), dict(
  DESCRIPTOR = _HEARTBEAT,
  __module__ = 'nrpc.nrpc_pb2'
  # @@protoc_insertion_point(class_scope:nrpc.HeartBeat)
  ))
_sym_db.RegisterMessage(HeartBeat)

google_dot_protobuf_dot_descriptor__pb2.FileOptions.RegisterExtension(packageSubject)
google_dot_protobuf_dot_descriptor__pb2.FileOptions.RegisterExtension(packageSubjectParams)
serviceSubjectRule.enum_type = _SUBJECTRULE
google_dot_protobuf_dot_descriptor__pb2.FileOptions.RegisterExtension(serviceSubjectRule)
methodSubjectRule.enum_type = _SUBJECTRULE
google_dot_protobuf_dot_descriptor__pb2.FileOptions.RegisterExtension(methodSubjectRule)
google_dot_protobuf_dot_descriptor__pb2.ServiceOptions.RegisterExtension(serviceSubject)
google_dot_protobuf_dot_descriptor__pb2.ServiceOptions.RegisterExtension(serviceSubjectParams)
google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(methodSubject)
google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(methodSubjectParams)
google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(streamedReply)

DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('Z\030github.com/nats-rpc/nrpc'))
# @@protoc_insertion_point(module_scope)



from .lib import (parse_subject, parse_subject_tail, streamed_reply_request,
                  streamed_reply_handler)
from .exc import ClientError

__all__ = [
    'ClientError',
    'parse_subject',
    'parse_subject_tail',
    'streamed_reply_request',
    'streamed_reply_handler',
]



import sys
import os

from . import nrpc_pb2 as nrpc
from .tmpl import template
from google.protobuf.compiler import plugin_pb2 as plugin

PY2 = sys.version_info.major == 2


class Generator:
    def __init__(self, request):
        self.request = request

    def lookup_fd(self, pkg):
        for fd in self.request.proto_file:
            if fd.package == pkg:
                return fd
        return None

    def get_fd_mod(self, fd):
        return os.path.splitext(fd.name)[0].replace("/", ".") + '_pb2'

    def get_mod_alias(self, mod):
        return mod.replace('_', '__').replace('.', '_dot_')

    def lookup_type(self, name):
        if not name.startswith('.'):
            raise ValueError(
                "lookup_type only accepts fully qualified names. Got '%s'" %
                name)
        path = name.lstrip('.').split('.')
        pkg = path
        fd = None
        while fd is None and len(pkg) > 1:
            pkg = pkg[:-1]
            fd = self.lookup_fd('.'.join(pkg))
        if fd is None:
            return None, None

        path = path[len(pkg):]

        return self.get_fd_mod(fd), '.'.join(path)

    def get_type(self, name):
        mod, tname = self.lookup_type(name)
        return '%s.%s' % (self.get_mod_alias(mod), tname)

    def extra_imports(self):
        pbmods = set()
        for sd in self.fd.service:
            for md in sd.method:
                for t in (md.input_type, md.output_type):
                    mod, _ = self.lookup_type(t)
                    pbmods.add(mod)
        return [(x, self.get_mod_alias(x)) for x in pbmods]

    def get_pkg_subject(self, fd):
        if fd.options.HasExtension(nrpc.packageSubject):
            return fd.options.Extensions[nrpc.packageSubject]
        return fd.package

    def get_pkg_params(self, fd):
        e = fd.options.Extensions[nrpc.packageSubjectParams]
        if e:
            return e
        return []

    def get_svc_subject(self, sd):
        s = sd.options.Extensions[nrpc.serviceSubject]
        if s:
            return s
        r = self.fd.options.Extensions[nrpc.serviceSubjectRule]
        if r == nrpc.COPY:
            return sd.name
        elif r == nrpc.TOLOWER:
            return sd.name.lower()
        raise Exception("Unknown subject rule %s" % r)

    def get_svc_params(self, sd):
        e = sd.options.Extensions[nrpc.serviceSubjectParams]
        if e is not None:
            return e
        return []

    def get_mt_subject(self, md):
        s = md.options.Extensions[nrpc.methodSubject]
        if s:
            return s
        r = self.fd.options.Extensions[nrpc.methodSubjectRule]
        if r == nrpc.COPY:
            return md.name
        elif r == nrpc.TOLOWER:
            return md.name.lower()
        raise Exception("Unknown subject rule %s" % r)

    def get_mt_params(self, md):
        e = md.options.Extensions[nrpc.methodSubjectParams]
        if e is not None:
            return e
        return []

    def mt_has_streamed_reply(self, md):
        s = md.options.Extensions[nrpc.streamedReply]
        if s:
            return s
        return False

    def get_fd(self, name):
        return next((fd for fd in self.request.proto_file if fd.name == name),
                    None)

    def generate(self):
        response = plugin.CodeGeneratorResponse()

        for f in self.request.file_to_generate:
            self.fd = self.get_fd(f)
            pbmod = os.path.splitext(self.fd.name)[0] + "_pb2"
            try:
                out_file = response.file.add()
                out_file.name = os.path.splitext(self.fd.name)[0] + "_nrpc.py"
                out_file.content = template.render(
                    g=self, fd=self.fd, pbmod=pbmod)
            finally:
                self.fd = None
        return response


def main():
    stdin = sys.stdin if PY2 else sys.stdin.buffer
    stdout = sys.stdout if PY2 else sys.stdout.buffer
    request = plugin.CodeGeneratorRequest.FromString(stdin.read())
    generator = Generator(request)
    response = generator.generate()
    stdout.write(response.SerializeToString())



from . import nrpc_pb2


class NrpcError(RuntimeError):
    type_ = None

    def __init__(self, message=None):
        self.message = message

    def as_nrpc_error(self):
        return nrpc_pb2.Error(type=self.type_, message=str(self))


class ClientError(NrpcError):
    type_ = nrpc_pb2.Error.CLIENT


class ServerError(NrpcError):
    type_ = nrpc_pb2.Error.SERVER


class EOS(NrpcError):
    type_ = nrpc_pb2.Error.EOS

    def __init__(self):
        self.message = None


class ServerTooBusy(NrpcError):
    type_ = nrpc_pb2.Error.SERVERTOOBUSY


def server_error(e):
    """Convert an exception into a nrpc.Error message"""
    return nrpc_pb2.Error(type=nrpc_pb2.Error.SERVER, message=str(e))


def from_error(error):
    """Convert a nrpc.Error message into a python exception"""
    if error.type == nrpc_pb2.Error.CLIENT:
        return ClientError(error.message)
    elif error.type == nrpc_pb2.Error.SERVER:
        return ServerError(error.message)
    elif error.type == nrpc_pb2.Error.EOS:
        return EOS()
    elif error.type == nrpc_pb2.Error.SERVERTOOBUSY:
        return ServerTooBusy(error.message)
    else:
        raise RuntimeError("Not an error: %s" % error)



import asyncio

from nats.aio.client import INBOX_PREFIX

import nrpc.exc
from . import nrpc_pb2


class InvalidSubject(Exception):
    pass


class ClientLost(RuntimeError):
    pass


class CanceledByClient(RuntimeError):
    pass


def parse_subject(
        package_subject,
        package_params_count,
        service_subject,
        service_params_count,
        subject,
):
    package_subject = package_subject.split('.') if package_subject else []
    service_subject = service_subject.split('.')

    minlen = len(package_subject) + package_params_count + \
        len(service_subject) + service_params_count + 1

    tokens = subject.split('.')

    if len(tokens) < minlen:
        raise InvalidSubject(
            "subject must contain %s tokens at least, got %s" % (minlen,
                                                                 subject))

    if tokens[:len(package_subject)] != package_subject:
        raise InvalidSubject(
            "subject should start with %s" % '.'.join(package_subject))

    tokens = tokens[len(package_subject):]

    package_params = tokens[:package_params_count]
    tokens = tokens[package_params_count:]

    if tokens[:len(service_subject)] != service_subject:
        raise InvalidSubject("subject should contain %s, got %s" %
                             ('.'.join(service_subject), subject))

    tokens = tokens[len(service_subject):]

    service_params = tokens[:service_params_count]
    tokens = tokens[service_params_count:]

    name = tokens[0]

    return package_params, service_params, name, tokens[1:]


def parse_subject_tail(method_params_count, tail):
    if len(tail) < method_params_count:
        raise InvalidSubject("subject tail is too short")
    method_params = tail[:method_params_count]
    tail = tail[method_params_count:]
    if len(tail) == 0:
        encoding = "protobuf"
    elif len(tail) == 1:
        encoding = tail[0]
    elif len(tail) > 1:
        raise InvalidSubject("subject tail is too long")
    return method_params, encoding


async def streamed_reply_request(nc, subject, req, timeout):
    if hasattr(nc, 'next_inbox'):
        inbox = nc.next_inbox()
    else:
        next_inbox = INBOX_PREFIX[:]
        next_inbox.extend(nc._nuid.next())
        inbox = next_inbox.decode()

    heartbeat_subject = inbox + ".heartbeat"

    msg_queue = asyncio.Queue(1)

    async def handler(msg):
        await msg_queue.put(msg)

    async def heartbeat():
        while True:
            await asyncio.sleep(1)
            await nc.publish(heartbeat_subject,
                             nrpc_pb2.HeartBeat().SerializeToString())

    heartbeat_task = asyncio.ensure_future(heartbeat())

    sid = await nc.subscribe(inbox, cb=handler)

    try:
        await nc.publish_request(subject, inbox, req)

        while True:
            msg = await asyncio.wait_for(msg_queue.get(), timeout)

            # Is it a keep-alive msg ?
            if len(msg.data) == 1 and msg.data[0] == 0:
                continue

            # Is it an error ?
            if len(msg.data) > 1 and msg.data[0] == 0:
                err = nrpc_pb2.Error.FromString(msg.data[1:])
                if err.type == nrpc_pb2.Error.EOS:
                    return
                raise Error

            yield msg
    finally:
        heartbeat_task.cancel()
        await nc.unsubscribe(sid)


async def heartbeat_listener(nc, subject, oncancel):
    queue = asyncio.Queue(1)

    def handler(msg):
        queue.put(msg)

    sid = await nc.subscribe(nc, subject, handler)

    try:
        while True:
            try:
                msg = await asyncio.wait_for(queue.get(), 2)
                beat = nrpc_pb2.HeartBeat.FromString(msg.data)
                if beat.last:
                    oncancel()
                    return
            except asyncio.TimeoutError as e:
                oncancel()
                return
    finally:
        await nc.unsubscribe(sid)


async def wrap_gen(nc, inbox, async_gen):
    while True:
        try:
            reply = await asyncio.wait_for(async_gen.__anext__(), 1)
            data = reply.SerializeToString()
            await nc.publish(inbox, data)
        except asyncio.TimeoutError:
            # Send a keep-alive message
            await nc.publish(inbox, b'\0')
        except StopAsyncIteration:
            eos = nrpc_pb2.Error()
            eos.type = nrpc_pb2.Error.EOS
            await nc.publish(inbox, b'\0' + eos.SerializeToString())
            return


async def streamed_reply_handler(nc, inbox, async_gen):
    heartbeat_subject = inbox + ".heartbeat"

    err = None

    queue = asyncio.Queue(1)

    EOQ = object()

    async def consume_generator():
        try:
            async for reply in async_gen:
                await queue.put(reply)
        except nrpc.ClientError as e:
            await queue.put(e)
        except asyncio.CancelledError as e:
            await queue.put(e)
        except Exception as e:
            await queue.put(nrpc.exc.server_error(e))

        await queue.put(EOQ)

    task = asyncio.ensure_future(consume_generator())

    heartbeat_task = asyncio.ensure_future(
        heartbeat_listener(nc, heartbeat_subject, task.cancel))

    msgCount = 0

    try:
        while True:
            try:
                reply = await asyncio.wait_for(queue.get(), 1)
                if isinstance(reply, asyncio.CancelledError):
                    return
                if reply is EOQ:
                    reply = nrpc_pb2.Error()
                    reply.type = nrpc_pb2.Error.EOS
                    reply.msgCount = msgCount
                if isinstance(reply, nrpc.exc.NrpcError):
                    reply = reply.as_nrpc_error()
                data = reply.SerializeToString()
                if isinstance(reply, nrpc_pb2.Error):
                    data = b'\x00' + data
                else:
                    msgCount += 1
                await nc.publish(inbox, data)
                if isinstance(reply, nrpc_pb2.Error):
                    return
            except asyncio.TimeoutError:
                await nc.publish(inbox, b'\x00')

    finally:
        task.cancel()
        heartbeat_task.cancel()



import distutils.cmd
import subprocess

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()


class protoc(distutils.cmd.Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.check_call(
            [
                "protoc",
                "--proto_path",
                ".",
                "--proto_path",
                "nrpc",
                "--python_out",
                ".",
                "nrpc/nrpc.proto",
            ]
        )


setup(
    name="python-nrpc",
    version="0.0.2",
    description="A python code generator and lib for Nats RPC",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["nrpc"],
    cmdclass={"protoc": protoc},
    entry_points={"console_scripts": ["protoc-gen-pynrpc=nrpc.gen:main"]},
    install_requires=["protobuf", "mako", "asyncio-nats-client"],
)
